#
# Use this to create valid ID Broker objects through the Provisioning API.
# Must be executed inside the ID Connector container on a school authority server.
#
# devsync -v performance-tests/ 10.200.?.?:/var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/
# univention-app shell ucsschool-id-connector
# cd /var/lib/univention-appcenter/apps/ucsschool-id-connector/conf/performance-tests/
# pip install ipython
# ipython
#
# import dev_tools
# kelvin_auth = dev_tools.KelvinCredentials(
#     username="Administrator", password="univention", host="provisioning.uni.dtr", verify=False
# )
# await dev_tools.create_provisioning_admin_users(kelvin_auth)
# await dev_tools.create_x_schools_in_y_school_authorities(4, 20, "provisioning.uni.dtr")
# await dev_tools.create_z_classes_in_x_schools_in_y_school_authorities(5, 4, 20, "provisioning.uni.dtr")
# await dev_tools.add_x_users_to_all_classes_of_y_school_authorities(2, 20, "provisioning.uni.dtr", kelvin_auth)
#
import asyncio
import random
import re
import uuid
from typing import Dict, Iterator, List, TypeVar, Union

import httpx
from faker import Faker
from pydantic import BaseModel

from ucsschool.kelvin.client import (
    KelvinClientError,
    School as KelvinSchool,
    SchoolClass as KelvinSchoolClass,
    SchoolClassResource as KelvinSchoolClassResource,
    SchoolResource as KelvinSchoolResource,
    Session as KelvinSession,
    User as KelvinUser,
    UserResource as KelvinUserResource,
)
from ucsschool_id_connector import plugin_loader
from ucsschool_id_connector.models import SchoolAuthorityConfiguration

fake = Faker()
plugin_loader.load_plugins()

from idbroker.id_broker_client import (  # noqa: E402
    IDBrokerError,
    IDBrokerSchool,
    IDBrokerSchoolClass,
    IDBrokerUser,
    School,
    SchoolClass,
    SchoolContext,
    User,
)

T = TypeVar("T")
JSON_FILE_PATTERN = re.compile(r"(?P<sa>.+?)-(?P<ou>.+)_(?P<type>.+).json")


class KelvinOutOfBandClientError(Exception):
    ...


class KelvinCredentials(BaseModel):
    username: str
    password: str
    host: str
    verify: Union[bool, str]


class KelvinOutOfBandClient:
    def __init__(self, auth: KelvinCredentials):
        self.auth = auth

    async def schools_get(self, *, name: str) -> List[KelvinSchool]:
        try:
            async with KelvinSession(**self.auth.dict()) as session:
                return [
                    school async for school in KelvinSchoolResource(session=session).search(name=name)
                ]
        except (KelvinClientError, httpx.HTTPError) as exc:
            raise KelvinOutOfBandClientError(
                f"Error retrieving schools using {self.auth.username!r} with name={name!r}: {exc!s}"
            ) from exc

    async def school_classes_get(self, *, name: str, school: str) -> List[KelvinSchoolClass]:
        try:
            async with KelvinSession(**self.auth.dict()) as session:
                return [
                    school_class
                    async for school_class in KelvinSchoolClassResource(session=session).search(
                        name=name, school=school
                    )
                ]
        except (KelvinClientError, httpx.HTTPError) as exc:
            raise KelvinOutOfBandClientError(
                f"Error retrieving school classes using {self.auth.username!r} with name={name!r} and "
                f"school={school!r}: {exc!s}"
            ) from exc


def split_list(li: List[T], chunk_size: int) -> Iterator[List[T]]:
    """Split a list into chunks of equal size (except last chunk, which is <= `chunk_size`)."""
    len_li = len(li)
    for i in range(0, len_li, chunk_size):
        yield li[i : i + chunk_size]


async def create_provisioning_admin_users(kelvin_auth: KelvinCredentials) -> Dict[str, str]:
    """
    Create 20 users (through Kelvin) that can be used to communicate with the provisioning API.

    usernames: `provisioning-TraegerPerf01` to `provisioning-TraegerPerf20`
    passwords: `Univention.Provisioning.01` to `Univention.Provisioning.20`.

    :returns mapping from username to password
    """
    print("Creating 20 provisioning-TraegerPerf?? users...")
    expected_usernames = {f"provisioning-TraegerPerf{num:02d}" for num in range(1, 21)}
    # username=auth.username, password=auth.password, host=auth.host, verify=auth.verify
    async with KelvinSession(**kelvin_auth.dict()) as session:
        users = [
            user
            async for user in KelvinUserResource(session=session).search(
                name="provisioning-TraegerPerf*"
            )
        ]
        existing_usernames = {user.name for user in users if user.name in expected_usernames}
        missing_usernames = expected_usernames - existing_usernames
        for username in sorted(missing_usernames):
            print(f"Creating provisioning admin user {username!r}...")
            user = KelvinUser(
                school="DEMOSCHOOL",
                schools=["DEMOSCHOOL"],
                roles=["staff"],
                name=username,
                firstname="provisioning-TraegerPerf",
                lastname=f"number{username[-2:]}",
                password=f"Univention.Provisioning.{username[-2:]}",
                record_uid=username,
                source_uid="provisioning-TraegerPerf",
                session=session,
            )
            await user.save()
            users.append(user)
    print("... done.")
    return {user.name: f"Univention.Provisioning.{user.name[-2:]}" for user in users}


def create_school_authority_configuration(
    school_authority_name: str, password: str, provisioning_host: str
) -> SchoolAuthorityConfiguration:
    return SchoolAuthorityConfiguration(
        name=school_authority_name,
        url=f"https://{provisioning_host}/",
        plugin_configs={
            "id_broker": {
                "username": f"provisioning-{school_authority_name}",
                "password": password,
                "version": 1,
            }
        },
        active=True,
        plugins=["id_broker-users", "id_broker-groups"],
    )


async def create_school_through_provisioning_api(
    school_authority: SchoolAuthorityConfiguration, school_name: str, entry_uuid: str = None
) -> School:
    IDBrokerSchool._share_token = False
    id_broker_school = IDBrokerSchool(school_authority, "id_broker")
    id_broker_school.configuration.verify_ssl = False
    entry_uuid = entry_uuid or str(uuid.uuid4())
    school = School(id=entry_uuid, name=school_name, display_name="Performance Test perm OU")
    print(f"  Creating school {school.name!r}...")
    await id_broker_school.create(school)
    return school


async def create_school_class_through_provisioning_api(
    school_authority: SchoolAuthorityConfiguration,
    school_name: str,
    class_name: str,
    entry_uuid: str = None,
) -> SchoolClass:
    IDBrokerSchoolClass._share_token = False
    id_broker_class = IDBrokerSchoolClass(school_authority, "id_broker")
    id_broker_class.configuration.verify_ssl = False
    entry_uuid = entry_uuid or str(uuid.uuid4())
    school_class = SchoolClass(
        id=entry_uuid,
        name=class_name,
        description="Performance Test perm class",
        school=school_name,
        members=[],
    )
    await id_broker_class.create(school_class)
    return school_class


async def create_user_through_provisioning_api(
    kelvin_auth: KelvinCredentials,
    school_authority: SchoolAuthorityConfiguration,
    *,
    entry_uuid: str = None,
    first_name: str = None,
    last_name: str = None,
    user_name: str = None,
    context: Dict[str, SchoolContext] = None,
) -> User:
    kelvin_client = KelvinOutOfBandClient(kelvin_auth)
    IDBrokerUser._share_token = False
    id_broker_user = IDBrokerUser(school_authority, "id_broker")
    id_broker_user.configuration.verify_ssl = False
    entry_uuid = entry_uuid or fake.unique.uuid4()
    first_name = first_name or fake.first_name()
    last_name = last_name or fake.last_name()
    user_name = user_name or f"{fake.user_name()[:12]}{fake.pyint(1000, 9999)}"
    if not context:
        schools = await kelvin_client.schools_get(name=f"{school_authority.name}-*")
        school = random.choice([s.name for s in schools])
        classes = await kelvin_client.school_classes_get(school=school, name="*")
        kls = random.choice([c.name for c in classes])
        role = random.choice(["student", "teacher"])
        context = {school: SchoolContext(classes=[kls], roles=[role])}
    user = User(
        id=entry_uuid,
        first_name=first_name,
        last_name=last_name,
        user_name=user_name,
        context=context,
    )
    await id_broker_user.create(user)
    return user


async def create_x_schools_in_y_school_authorities(
    num_schools: int, num_school_auths: int, provisioning_host: str
) -> None:
    for i in range(1, num_school_auths + 1):
        school_authority = create_school_authority_configuration(
            f"TraegerPerf{i:02}", f"Univention.Provisioning.{i:02}", provisioning_host
        )
        print(f"Creating {num_schools} schools in school authority {school_authority.name!r}...")
        for j in range(1, num_schools + 1):
            name = f"Schule{j:02}"
            print(f"    Creating school class {name!r}...")
            try:
                await create_school_through_provisioning_api(school_authority, name)
                print(f"  School {name!r} created.")
            except IDBrokerError as exc:
                if exc.status == 409:
                    print(f"  School {name!r} exists.")
                    continue
                raise


async def create_z_classes_in_x_schools_in_y_school_authorities(
    num_classes: int, num_schools: int, num_school_auths: int, provisioning_host: str
) -> None:
    """
    Run `create_x_schools_in_y_school_authorities(num_schools, num_school_auths, provisioning_host)`
    before this.
    """

    async def _create(s_a: SchoolAuthorityConfiguration, s_n: str, c_n: str):
        try:
            await create_school_class_through_provisioning_api(s_a, s_n, c_n)
            print(f"    School class {c_n!r} created.")
        except IDBrokerError as _exc:
            if _exc.status == 409:
                print(f"    School class {c_n!r} exists.")
            else:
                raise

    for i in range(1, num_school_auths + 1):
        school_authority = create_school_authority_configuration(
            f"TraegerPerf{i:02}", f"Univention.Provisioning.{i:02}", provisioning_host
        )
        print(f"Creating school classes in school authority {school_authority.name!r}...")
        for j in range(1, num_schools + 1):
            school_name = f"Schule{j:02}"
            print(f"  Creating {num_classes} school classes in school {school_name!r}...")
            class_names = [f"class{k:02}" for k in range(1, num_classes + 1)]
            for chunk in split_list(
                [_create(school_authority, school_name, name) for name in class_names], 4
            ):
                await asyncio.gather(*chunk)


async def add_x_users_to_all_classes_of_y_school_authorities(
    num_users_per_class: int,
    num_school_auths: int,
    provisioning_host: str,
    kelvin_auth: KelvinCredentials,
) -> None:
    kelvin_client = KelvinOutOfBandClient(kelvin_auth)
    for i in range(1, num_school_auths + 1):
        school_authority = create_school_authority_configuration(
            f"TraegerPerf{i:02}", f"Univention.Provisioning.{i:02}", provisioning_host
        )
        print(f"Checking classes of school authority {school_authority.name!r}...")
        schools = await kelvin_client.schools_get(name=f"{school_authority.name}-*")
        for school in schools:
            print(f"  Checking classes of school {school.name!r}...")
            classes = await kelvin_client.school_classes_get(school=school.name, name="*")
            for kls in classes:
                to_add = num_users_per_class - len(kls.users)
                if to_add > 0:
                    print(f"    Class {kls.name} has {len(kls.users)} users, adding {to_add}...")
                    # always add 1 teacher and the rest students
                    roles = ["teacher"] + ["student" for _ in range(to_add - 1)]
                    for role in roles:
                        school_name = school.name[len(f"{school_authority.name}-") :]
                        context = {school_name: SchoolContext(classes=[kls.name], roles=[role])}
                        try:
                            user: User = await create_user_through_provisioning_api(
                                kelvin_auth=kelvin_auth,
                                school_authority=school_authority,
                                context=context,
                            )
                            print(f"      User {user.user_name!r} created.")
                        except IDBrokerError as exc:
                            print(f"      Error creating user: {exc!s}")
